from typing import Union


class MissingValueError(ValueError):
    def __init__(self, params: Union[str, list]):
        super().__init__(f"Required value is missing. Provide one of: {params}")


class InvalidModelError(ValueError):
    def __init__(self):
        super().__init__("Invalid model provided.")


class InvalidValueError(ValueError):
    def __init__(self, params: list):
        super().__init__(f"Invalid value provided. Provide value of type: {params}")


class MissingModuleError(ImportError):
    def __init__(self, required_modules: dict):
        super().__init__(
            f"The following modules are needed to run this function: {', '.join(required_modules.keys())}. Please install them using: `pip install {' '.join(required_modules.values())}`"
        )



import os
from typing import Optional, Literal
from lyzr.base.llms import LLM, get_model
from lyzr.base.errors import MissingValueError


class VoiceBot:
    def __init__(
        self,
        api_key: Optional[str] = None,
        model_type: Optional[Literal["openai"]] = None,
        model_name: Optional[Literal["tts-1-hd", "tts-1"]] = None,
        model: Optional[LLM] = None,
        voice: Optional[
            Literal["echo", "alloy", "fable", "onyx", "nova", "shimmer"]
        ] = None,
    ):
        self.api_key = api_key or os.environ.get("OPENAI_API_KEY")
        if self.api_key is None:
            raise MissingValueError("API key")
        self.model = model or get_model(
            api_key=self.api_key,
            model_type=model_type or os.environ.get("MODEL_TYPE") or "openai",
            model_name=model_name or os.environ.get("MODEL_NAME") or "tts-1-hd",
            voice=voice or os.environ.get("VOICE") or "echo",
        )  # change get_model in lyzr.base.llms to accept **kwargs

    def text_to_speech(
        self,
        text: str,
    ):
        if self.model.model_name not in ["tts-1-hd", "tts-1"]:
            if self.model.model_type == "openai":
                self.model = get_model(
                    api_key=self.api_key,
                    model_type=self.model.model_type,
                    model_name="tts-1-hd",
                )
            else:
                raise ValueError(
                    "The text_to_speech function only works with the OpenAI's 'tts-1-hd' and 'tts-1' models."
                )
        response = self.model.run(input=text)
        # Save the synthesized speech to a file named "tts_output.mp3"
        response.stream_to_file("tts_output.mp3")

    def transcribe(self, audiofilepath):
        if self.model.model_name != "whisper-1":
            if self.model.model_type == "openai":
                self.model = get_model(
                    api_key=self.api_key,
                    model_type=self.model.model_type,
                    model_name="whisper-1",
                )
            else:
                raise ValueError(
                    "The transcribe function only works with the OpenAI's 'whisper-1' model."
                )

        response = self.model.run(audiofile=open(audiofilepath, "rb"))
        return response.text

    def text_to_notes(self, text):
        if self.model.model_name != "gpt-3.5-turbo":
            if self.model.model_type == "openai":
                self.model = get_model(
                    api_key=self.api_key,
                    model_type=self.model.model_type,
                    model_name="gpt-3.5-turbo",
                )
            else:
                raise ValueError(
                    "The text_to_notes function only works with the OpenAI's 'gpt-3.5-turbo' model."
                )

        # The system message acts as the prompt for the AI.
        system_message = "You are an expert in taking down notes as bullet points and summarizing big conversations. You make sure no detail is left out."

        # Format the user's message that will be sent to the model.
        user_message = f"Here is my conversation: {text}. Can you create bullet-point notes for this?"
        self.model.set_messages(
            model_prompts=[
                {"role": "system", "text": system_message},
                {"role": "user", "text": user_message},
            ]
        )
        # Use the LLM instance to communicate with OpenAI's API.
        response = self.model.run()

        # Parse the response to extract the notes.
        notes = response.choices[0].message.content

        return notes



import os
from openai import OpenAI
from typing import Optional, Literal
from lyzr.base.prompt import get_prompt_text
from lyzr.base.errors import MissingValueError, InvalidValueError


class LLM:
    def __init__(
        self,
        api_key: str,
        model_type: Optional[str] = None,
        model_name: Optional[str] = None,
        model_prompts: Optional[list[dict]] = None,
        voice: Optional[
            Literal["echo", "alloy", "fable", "onyx", "nova", "shimmer"]
        ] = None,  # Add support for specifying the voice model
        **kwargs,
    ):
        self.api_key = api_key
        self.model_type = model_type or "openai"
        self.model_name = model_name or "gpt-3.5-turbo"
        self.messages = None
        self.voice = voice or "nova"  # Default voice for TTS
        if model_prompts is not None:
            self.set_messages(model_prompts)
        for param in kwargs:
            setattr(self, param, kwargs[param])

    def set_messages(
        self,
        model_prompts: Optional[list[dict]] = None,
        messages: Optional[list[dict]] = None,
    ):
        if model_prompts is None and messages is None and self.messages is None:
            raise ValueError("Please set a value for the prompt")

        if messages is not None:
            self.messages = messages
            return self

        messages = []
        for prompt in model_prompts:
            messages.append(
                {"role": prompt["role"], "content": get_prompt_text(prompt)}
            )
        self.messages = messages
        return self

    def run(self, **kwargs):
        if self.api_key is None:
            raise ValueError(
                "Please provide an API key or set the API_KEY environment variable."
            )

        if self.messages is None:
            if "model_prompts" in kwargs:
                self.set_messages(model_prompts=kwargs["model_prompts"])
            elif "messages" in kwargs:
                self.set_messages(messages=[kwargs["messages"]])
            elif "input" in kwargs:
                self.input = kwargs["input"]
            elif "audiofile" in kwargs:
                self.audiofile = kwargs["audiofile"]
            else:
                raise MissingValueError(["model_prompts", "messages"])

        params = self.__dict__.copy()
        params.update(kwargs)
        for param in [
            "api_key",
            "model_prompts",
            "model_type",
            "model_name",
            "messages",
            "voice",
            "input",
            "audiofile",
        ]:
            if param in params:
                del params[param]

        # Instantiate the OpenAI client
        client = OpenAI(api_key=self.api_key)

        if self.model_type == "openai":
            # Check for Text-to-Speech models
            if self.model_name in ["tts-1", "tts-1-hd"]:
                if self.input is None:
                    raise MissingValueError("input")
                # Use 'voice' and any other options required by OpenAI's TTS endpoint
                response = client.audio.speech.create(
                    model=self.model_name,
                    voice=self.voice,
                    input=self.input,
                    **params,
                )
                return response

            # Check for transcription models like "whisper"
            elif self.model_name.startswith("whisper"):
                if self.audiofile is None:
                    raise MissingValueError("audiofile")
                # The transcription API may require a file-like object or binary data
                response = client.audio.transcriptions.create(
                    model=self.model_name,
                    file=self.audiofile,
                    **params,
                )
                return response

            # Else, handle chat completions
            else:
                completion = client.chat.completions.create(
                    model=self.model_name,
                    messages=self.messages,
                    **params,
                )
                return completion


def get_model(
    api_key: Optional[str] = None,
    model_type: Optional[str] = None,
    model_name: Optional[str] = None,
    **kwargs,
) -> LLM:
    return LLM(
        api_key=api_key or os.getenv("API_KEY"),
        model_type=model_type or os.getenv("MODEL_TYPE") or "openai",
        model_name=model_name or os.getenv("MODEL_NAME") or "gpt-3.5-turbo",
        **kwargs,
    )



import logging
from typing import List

import requests
from bs4 import BeautifulSoup
from llama_index.schema import Document
from tqdm import tqdm

from lyzr.utils.webpage_reader import LyzrWebPageReader

logger = logging.getLogger(__name__)


class LyzrWebsiteReader:
    def __init__(self):
        self.visited_links = set()

    @staticmethod
    def load_data(url: str) -> List[Document]:
        reqs = requests.get(url)
        soup = BeautifulSoup(reqs.text, "html.parser")

        all_urls = set()
        for link in soup.find_all("a"):
            href = link.get("href")
            if href is not None:
                all_urls.add(url + href)

        logger.info(f"Total URLs to process: {len(all_urls)}")
        web_reader = LyzrWebPageReader()
        documents = []
        for u in tqdm(all_urls, desc="Processing URLs"):
            documents.extend(web_reader.load_data(u))

        return documents



"""
With DataAnalyzr, you can analyze any dataframe 
derive actionable insights and create 
visually compelling and Intuitive visualizations.
"""
import os
import io
import re
import sys
import shutil
import warnings
import datetime
from pathlib import Path
from types import TracebackType
from contextlib import AbstractContextManager
from typing import Optional, Union, Type

import numpy as np
from PIL import Image
import pandas as pd
from pandas.errors import EmptyDataError

from lyzr.base.prompt import Prompt
from lyzr.base.llms import LLM, get_model
from lyzr.base.errors import MissingValueError
from lyzr.base.file_utils import read_file

PATTERN_PYTHON_CODE_BLOCK = r"```python\n(.*?)\n```"

warnings.filterwarnings("ignore")


class CapturePrints(AbstractContextManager):
    """
    A context manager for capturing output to stdout.

    This class can be used to capture prints and other output
    that normally go to stdout. It can be useful whenever
    output to stdout needs to be captured or suppressed.
    """

    def __enter__(self) -> "CapturePrints":
        self._old_stdout = sys.stdout
        sys.stdout = self._mystdout = io.StringIO()
        return self

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[Exception],
        exc_tb: Optional[TracebackType],
    ) -> Optional[bool]:
        sys.stdout = self._old_stdout
        # Can decide to handle exceptions here or let them propagate
        return None  # Returning None means any exceptions will propagate

    def get_value(self) -> str:
        """Retrieve the captured stdout content."""
        return self._mystdout.getvalue()


class DataAnalyzr:
    """The DataAnalyzr Class for analyzing dataframes."""

    def __init__(
        self,
        df: Union[str, pd.DataFrame] = None,
        api_key: Optional[str] = None,
        model: Optional[LLM] = None,
        model_type: Optional[str] = None,
        model_name: Optional[str] = None,
        user_input: Optional[str] = None,
        seed: int = None,
    ):
        self.model = (
            model
            or os.environ.get("LLM_MODEL")
            or get_model(
                api_key=api_key or os.environ.get("OPENAI_API_KEY"),
                model_type=model_type or os.environ.get("MODEL_TYPE", "openai"),
                model_name=model_name or os.environ.get("MODEL_NAME", "gpt-3.5-turbo"),
            )
        )
        # Process the dataframe parameter
        if df is None:
            df = os.environ.get(
                "DATAFRAME"
            )  # Try getting from environment if not provided
        if df is None:
            # If df is still None, neither a DataFrame nor a valid path were provided
            raise MissingValueError(["dataframe"])
        elif isinstance(df, str):
            self.df = self._clean_df(read_file(df))
        elif isinstance(df, pd.DataFrame):
            # You might want to check whether df is empty
            if df.empty:
                raise EmptyDataError("The provided DataFrame is empty.")
            self.df = self._clean_df(df)
        else:
            raise ValueError("df must be a path to a file or a pd.DataFrame object.")

        if isinstance(df, str):
            self.df = self._clean_df(read_file(df))

        self.user_input = user_input
        self.seed = seed

    def _clean_df(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Clean the dataframe.

        Parameters:
        - df (pd.DataFrame) : Input Dataframe to be cleaned.

        Returns:
        - pd.DataFrame: cleaned dataframe with no columns having >50% missing values,
        no unnamed columns, no duplicates, and NaNs replaced in numerical/categorical columns.
        """
        # Removing columns having more than 50% of missing values
        df = df[df.columns[df.isnull().mean() < 0.5]]

        # Getting the columns which are categorical
        cat_columns = df.select_dtypes(include=["object"]).columns

        # Getting the columns which are numerical
        num_columns = df.select_dtypes(include=[np.number]).columns

        # Replacing missing categorical values with the most frequent value(mode)
        df[cat_columns] = df[cat_columns].apply(lambda x: x.fillna(x.mode()[0]))

        # Replacing missing numerical values with the mean
        df[num_columns] = df[num_columns].apply(lambda x: x.fillna(x.mean()))

        # Removing "Unnamed:" columns if any
        df = df.loc[:, ~df.columns.str.contains("^Unnamed")]

        # Removing duplicates
        df = df.drop_duplicates(keep="first")

        return df

    def _get_analysis_steps(self, user_input: str = None) -> str:
        """
        Get the steps to perform the analysis.

        Parameters:
        - user_input (str):
            The user input based on which analysis steps are generated. (default: None)

        Raises:
        - MissingValueError: If user input is missing.

        Returns:
        - str: The analysis steps based on user input and dataframe.
        """
        self.user_input = user_input or self.user_input
        if self.user_input is None:
            raise MissingValueError(["user_input"])

        self.model.set_messages(
            [
                {
                    "prompt": Prompt("system_analysis_steps_pt"),
                    "role": "system",
                },
                {
                    "prompt": Prompt("analysis_steps_pt").format(
                        user_input=self.user_input,
                        df_head=self.df.head(5),
                        df_columns=self.df.columns.tolist(),
                    ),
                    "role": "user",
                },
            ]
        )

        steps = self.model.run(temperature=0.1).choices[0].message.content

        return steps

    def _get_analysis_code(self, instructions: str, user_input: str = None) -> str:
        """
        Get the Python code to perform the analysis.

        Parameters:
        - instructions (str): The instructions to perform the analysis.
        - user_input (str):
            The user input based on which analysis code is generated. (default: None)

        Raises:
        - MissingValueError: If user input is missing.

        Returns:
        - str: The analysis Python code based on user input and dataframe.
        """
        self.user_input = user_input or self.user_input
        if self.user_input is None:
            raise MissingValueError(["user_input"])

        self.model.set_messages(
            [
                {
                    "prompt": Prompt("system_code_pt"),
                    "role": "system",
                },
                {
                    "prompt": Prompt("analysis_code_pt").format(
                        user_input=self.user_input,
                        instructions=instructions,
                        df_head=self.df.head(5),
                        df_columns=self.df.columns.tolist(),
                    ),
                    "role": "user",
                },
            ]
        )
        model_response = self.model.run(temperature=0.1).choices[0].message.content

        python_code_blocks = re.findall(
            PATTERN_PYTHON_CODE_BLOCK, model_response, re.DOTALL
        )

        try:
            python_code = python_code_blocks[0]
        except IndexError:
            python_code = model_response

        return python_code

    def _get_visualization_steps(self, user_input: str = None) -> str:
        """
        Get the steps to perform the visualization.

        Parameters:
        - user_input (str):
            The user input based on which visualization steps are generated. (default: None)

        Raises:
        - MissingValueError: If user input is missing.

        Returns:
        - str: The visualization steps based on user input and dataframe.
        """
        self.user_input = user_input or self.user_input
        if self.user_input is None:
            raise MissingValueError(["user_input"])

        self.model.set_messages(
            [
                {
                    "prompt": Prompt("system_visualization_pt"),
                    "role": "system",
                },
                {
                    "prompt": Prompt("visualization_steps_pt").format(
                        user_input=self.user_input,
                        df_head=self.df.head(5),
                        df_columns=self.df.columns.tolist(),
                    ),
                    "role": "user",
                },
            ]
        )

        steps = self.model.run(temperature=0.1).choices[0].message.content

        return steps

    def _get_visualiztion_code(self, instructions: str, user_input: str = None) -> str:
        """
        Get the Python code to generate the visualization.

        Parameters:
        - instructions (str): The instructions to generate the visualization.
        - user_input (str): The user input based on which visualization code is generated. (default: None)

        Raises:
        - MissingValueError: If user input is missing.

        Returns:
        - str: The visualization Python code based on user input and dataframe.
        """
        self.user_input = user_input or self.user_input
        if self.user_input is None:
            raise MissingValueError(["user_input"])

        self.model.set_messages(
            [
                {
                    "prompt": Prompt("system_vis_code_pt"),
                    "role": "system",
                },
                {
                    "prompt": Prompt("visualization_code_pt").format(
                        user_input=self.user_input,
                        df_head=self.df.head(5),
                        df_columns=self.df.columns.tolist(),
                        instructions=instructions,
                    ),
                    "role": "user",
                },
            ]
        )

        model_response = self.model.run(temperature=0.1).choices[0].message.content

        python_code_blocks = re.findall(
            PATTERN_PYTHON_CODE_BLOCK, model_response, re.DOTALL
        )

        try:
            python_code = python_code_blocks[0]
        except IndexError:
            python_code = model_response

        return python_code

    def _correct_code(self, python_code: str, error_message: str) -> str:
        """
        Correct the Python code based on the error message.

        Parameters:
        - python_code (str): The Python code to be corrected.
        - error_message (str): The error message based on which the code is corrected.

        Returns:
        - str: The corrected Python code.
        """
        corrected_python_code = ""

        self.model.set_messages(
            [
                {
                    "prompt": Prompt("system_correct_code_pt"),
                    "role": "system",
                },
                {
                    "prompt": Prompt("correct_code_pt").format(
                        python_code=python_code,
                        error_message=error_message,
                        df_columns=self.df.columns.tolist(),
                        df_head=self.df.head(5),
                        user_input=self.user_input,
                    ),
                    "role": "user",
                },
            ]
        )
        model_response = self.model.run(temperature=0.1).choices[0].message.content

        python_code_blocks = re.findall(
            PATTERN_PYTHON_CODE_BLOCK, model_response, re.DOTALL
        )

        try:
            corrected_python_code = python_code_blocks[0]
        except IndexError:
            corrected_python_code = model_response

        return corrected_python_code

    def _get_analysis_output(self, user_input: str = None) -> str:
        """
        Get the output of the analysis.

        Parameters:
        - user_input (str):
            The user input based on which analysis output is generated. (default: None)

        Raises:
        - MissingValueError: If user input is missing.

        Returns:
        - str: The analysis output after running the analysis python code.
        """
        self.user_input = user_input or self.user_input
        if self.user_input is None:
            raise MissingValueError(["user_input"])
        steps = self._get_analysis_steps()

        analysis_python_code = self._get_analysis_code(steps)

        with CapturePrints() as c:
            exec_scope = {"df": self.df}
            try:
                exec(analysis_python_code, exec_scope)
            except Exception as e:
                error_message = str(e)
                print(f"Error. {type(e).__name__}: {error_message}")
                analysis_python_code = self._correct_code(
                    analysis_python_code, error_message
                )
                exec(analysis_python_code, exec_scope)

        output = c.get_value()
        return output

    def _load_images_in_current_directory(self) -> dict[str, bytes]:
        """
        Load all PNG images in the current directory and convert them to a dictionary

        Returns:
        - Dict[str, bytes]:
            A dictionary where the key is the filename and the value is byte data of the image.
        """
        current_directory = os.getcwd()
        image_data = {}

        for filename in os.listdir(current_directory):
            if filename.lower().endswith(".png"):
                image_path = os.path.join(current_directory, filename)
                with Image.open(image_path) as img:
                    byte_arr = io.BytesIO()
                    img.save(byte_arr, format="PNG")
                    image_data[filename] = byte_arr.getvalue()

        return image_data

    def _move_visualization_files(
        self, source: Path, destination: Path, file_type: str
    ) -> None:
        """
        Move generated visualization files of a given type,
        from a source directory to a destination directory.

        Parameters:
        - source (Path): The source directory.
        - destination (Path): The destination directory.
        - file_type (str): The type of files to be moved.

        Post-condition:
        - The files of the specified type are moved from source to destination.
        - The files are renamed to include the current datetime in their names for uniqueness.
        """
        if not destination.exists():
            destination.mkdir(parents=True)
        for file_name in source.glob(f"*.{file_type}"):
            base = file_name.stem
            now = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
            new_file_name = f"{base}-{now}.{file_type}"
            new_file_path = destination / new_file_name
            shutil.move(str(file_name), str(new_file_path))

    def analysis_insights(self, user_input: str = None) -> str:
        """
        Get insights from the analysis in three bullet points.

        Parameters:
        - user_input (str): The user input based on which insights are generated. (default: None)

        Raises:
        - MissingValueError: If user input is missing.

        Returns:
        - str: The analysis insights derived from the dataframe based on user input.
        """
        self.user_input = user_input or self.user_input
        if self.user_input is None:
            raise MissingValueError(["user_input"])

        analysis_output = self._get_analysis_output()

        if len(analysis_output) > 6000:
            analysis_output = analysis_output[:3000] + "..." + analysis_output[-3000:]

        prompt_messages = [
            {
                "prompt": Prompt("system_analysis_pt"),
                "role": "system",
            },
            {
                "prompt": Prompt("analysis_output_pt").format(
                    user_input=self.user_input, analysis_output=analysis_output
                ),
                "role": "user",
            },
        ]

        self.model.set_messages(prompt_messages)

        analysis = (
            self.model.run(
                temperature=0.3, top_p=1, frequency_penalty=0, presence_penalty=0
            )
            .choices[0]
            .message.content
        )

        return analysis

    def visualizations(
        self,
        user_input: str = None,
        dir_path: Path = Path("./generated_plots"),
    ) -> list[Image.Image]:
        """
        Get visualizations of the analysis and
        save the generated plot images to the specified directory.

        Parameters:
        - user_input (str):
            The user input data based on which visualizations are created. (default: None)
        - dir_path (Path):
            The directory path to save the visualizations (default: "./generated_plots")

        Raises:
        - MissingValueError: If user input is not provided.

        Returns:
        - List[Image.Image]: A list of PIL Image objects representing the saved visualizations.
        """
        self.user_input = user_input or self.user_input
        if self.user_input is None:
            raise MissingValueError(["user_input"])
        visualization_steps = self._get_visualization_steps()
        visualization_python_code = self._get_visualiztion_code(visualization_steps)

        exec_scope = {"df": self.df}
        try:
            exec(visualization_python_code, exec_scope)
        except Exception as e:
            error_message = str(e)
            visualization_python_code = self._correct_code(
                visualization_python_code, error_message
            )
            exec(visualization_python_code, exec_scope)

        image_list = self._load_images_in_current_directory()
        self._move_visualization_files(Path("./"), dir_path, "png")
        return image_list

    def dataset_description(self) -> str:
        """
        Generate a brief description of the dataset currently in use.

        Returns:
        - str: A string providing a description of the dataset.
        """
        self.model.set_messages(
            [
                {
                    "prompt": Prompt("system_describe_dataset_pt"),
                    "role": "system",
                },
                {
                    "prompt": Prompt("describe_dataset_pt").format(
                        df_head=self.df.head(5),
                    ),
                    "role": "user",
                },
            ]
        )

        description = (
            self.model.run(
                temperature=1, top_p=0.3, frequency_penalty=0.7, presence_penalty=0.3
            )
            .choices[0]
            .message.content
        )

        return description

    def ai_queries_df(self, dataset_description: Optional[str] = None) -> str:
        """
        Returns AI-generated queries for data analysis related to the dataset.

        Parameters:
        - dataset_description (str, optional):
            A description of the dataset. If not provided, it will be generated.

        Returns:
        - str: Queries for data analysis related to the dataset.
        """
        if dataset_description is None:
            dataset_description = self.dataset_description()

        self.model.set_messages(
            [
                {
                    "prompt": Prompt("system_ai_queries_pt"),
                    "role": "system",
                },
                {
                    "prompt": Prompt("ai_queries_pt").format(
                        dataset_description=dataset_description,
                        df_head=self.df.head(5),
                    ),
                    "role": "user",
                },
            ]
        )

        ai_queries_df = (
            self.model.run(
                temperature=1, top_p=0.3, frequency_penalty=0.7, presence_penalty=0.3
            )
            .choices[0]
            .message.content
        )

        return ai_queries_df

    def analysis_recommendation(
        self, user_input: Optional[str] = None, number_of_recommendations: int = 4
    ) -> str:
        """
        Get recommendation on what analysis to perform on the dataset.

        Parameters:
        - user_input (str, optional): The user input data. Defaults to None.
        - number_of_recommendations (int, optional):
            The number of recommendations to return. Defaults to 4.

        Returns:
        - str: Recommendations for analysis.
        """
        formatted_user_input: str = (
            Prompt("formatted_user_input_pt").format(user_input=user_input)
            if user_input is not None
            else ""
        )

        self.model.set_messages(
            [
                {
                    "prompt": Prompt("system_recommendations_pt").format(
                        number_of_recommendations=number_of_recommendations,
                    ),
                    "role": "system",
                },
                {
                    "prompt": Prompt("analysis_recommendations_pt").format(
                        number_of_recommendations=number_of_recommendations,
                        df_head=self.df.head(5),
                        df_columns=self.df.columns.tolist(),
                        formatted_user_input=formatted_user_input,
                    ),
                    "role": "system",
                },
            ]
        )

        recommendations = (
            self.model.run(temperature=0.2, seed=self.seed).choices[0].message.content
        )
        return recommendations

    def recommendations(
        self,
        insights: Optional[str] = None,
        user_input: Optional[str] = None,
        schema: Optional[list] = None,
    ) -> str:
        """
        Get recommendations based on the analysis insights.

        Parameters:
        - insights (str, optional): The analysis insights. Defaults to None.
        - user_input (str, optional): The user input. Defaults to None.
        - schema (list, optional):
            The schema for the recommendations. Defaults to a predefined schema.

        Raises:
        - MissingValueError: If user_input is not provided.

        Returns:
        - str: Recommendations as a string.
        """
        self.user_input = user_input or self.user_input

        if self.user_input is None:
            raise MissingValueError(["user_input"])

        if insights is None:
            insights = self.analysis_insights(user_input=user_input)

        schema = schema or [
            {
                "Recommendation": "string",
                "Basis of the Recommendation": "string",
                "Impact if implemented": "string",
            },
            {
                "Recommendation": "string",
                "Basis of the Recommendation": "string",
                "Impact if implemented": "string",
            },
            {
                "Recommendation": "string",
                "Basis of the Recommendation": "string",
                "Impact if implemented": "string",
            },
        ]

        self.model.set_messages(
            [
                {
                    "prompt": Prompt("system_recommendation_pt").format(schema=schema),
                    "role": "system",
                },
                {
                    "prompt": Prompt("recommendation_pt").format(
                        user_input=user_input, insights=insights
                    ),
                    "role": "user",
                },
            ]
        )

        recommendations = (
            self.model.run(
                temperature=1, top_p=0.3, frequency_penalty=0.7, presence_penalty=0.3
            )
            .choices[0]
            .message.content
        )

        return recommendations

    def tasks(
        self,
        user_input: Optional[str] = None,
        insights: Optional[str] = None,
        recommendations: Optional[str] = None,
    ) -> str:
        """
        Generate tasks based on the given user input, analysis insights, and recommendations.

        Parameters:
        - user_input (Optional[str]):
            The user input data. If None, the class's user_input attribute is used.
        - insights (Optional[str]):
            The analysis insights. If None, generated by analysis_insights method.
        - recommendations (Optional[str]):
            The analysis recommendations. If None, generated by recommendations method.

        Raises:
        - MissingValueError: If user_input is not provided.

        Returns:
        - str: The generated tasks.
        """
        self.user_input = user_input or self.user_input
        if self.user_input is None:
            raise MissingValueError(["user_input"])

        if insights is None:
            insights = self.analysis_insights(user_input=user_input)

        if recommendations is None:
            recommendations = self.recommendations(
                insights=insights, user_input=user_input
            )

        self.model.set_messages(
            [
                {
                    "prompt": Prompt("system_tasks_pt"),
                    "role": "system",
                },
                {
                    "prompt": Prompt("task_pt").format(
                        user_input=user_input,
                        insights=insights,
                        recommendations=recommendations,
                    ),
                    "role": "user",
                },
            ]
        )

        tasks = (
            self.model.run(
                temperature=1, top_p=0.3, frequency_penalty=0.7, presence_penalty=0.3
            )
            .choices[0]
            .message.content
        )

        return tasks



import logging
from typing import Optional, Union

from llama_index import ServiceContext
from llama_index.callbacks import CallbackManager
from llama_index.embeddings.utils import EmbedType
from llama_index.llms.utils import LLMType
from llama_index.prompts import PromptTemplate
from llama_index.prompts.base import BasePromptTemplate
from llama_index.node_parser import (
    SimpleNodeParser,
)

logger = logging.getLogger(__name__)


class LyzrService:
    @staticmethod
    def from_defaults(
        llm: Optional[LLMType] = "default",
        embed_model: Optional[EmbedType] = "default",
        system_prompt: str = None,
        query_wrapper_prompt: Union[str, BasePromptTemplate] = None,
        **kwargs,
    ) -> ServiceContext: 
        if isinstance(query_wrapper_prompt, str):
            query_wrapper_prompt = PromptTemplate(template=query_wrapper_prompt)

        callback_manager: CallbackManager = kwargs.get(
            "callback_manager", CallbackManager()
        )

        node_parser = SimpleNodeParser.from_defaults(
            chunk_size=750,
            chunk_overlap=100,
            callback_manager=callback_manager,
        )

        service_context = ServiceContext.from_defaults(
            llm=llm,
            embed_model=embed_model,
            system_prompt=system_prompt,
            query_wrapper_prompt=query_wrapper_prompt,
            callback_manager=callback_manager,
            node_parser=node_parser,
            **kwargs,
        )

        return service_context



from typing import List

from langchain.document_loaders import PDFMinerLoader
from llama_index.readers.base import BaseReader
from llama_index.schema import Document


class LyzrPDFReader(BaseReader):
    def __init__(self) -> None:
        try:
            from pdfminer.high_level import extract_text 
        except ImportError:
            raise ImportError(
                "`pdfminer` package not found, please install it with "
                "`pip install pdfminer.six`"
            )

    def load_data(self, file_path: str, extra_info: dict = None) -> List[Document]:
        loader = PDFMinerLoader(str(file_path))
        langchain_documents = loader.load()  

        documents = []
        for langchain_document in langchain_documents:
            doc = Document.from_langchain_format(langchain_document)

            if extra_info is not None:
                doc.metadata.update(extra_info)

            documents.append(doc)

        return documents
